use {
    super::{Rapid, SemVer, Version},
    serde::{Deserialize, Serialize},
    std::{cmp::Ordering, error::Error, fmt, num::ParseIntError, str::FromStr},
};

#[derive(Clone, Debug, Serialize, Deserialize, PartialEq, Eq)]
pub struct Extended {
    pub major: u32,
    pub minor: u32,
    pub patch: u32,
    pub ext: u32,
}

impl fmt::Display for Extended {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}.{}.{}.{}",
            self.major, self.minor, self.patch, self.ext
        )
    }
}

impl PartialOrd for Extended {
    fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
        match self.major.partial_cmp(&other.major) {
            Some(Ordering::Greater) => Some(Ordering::Greater),
            Some(Ordering::Less) => Some(Ordering::Less),
            None => None,
            Some(Ordering::Equal) => match self.minor.partial_cmp(&other.minor) {
                Some(Ordering::Greater) => Some(Ordering::Greater),
                Some(Ordering::Less) => Some(Ordering::Less),
                None => None,
                Some(Ordering::Equal) => match self.patch.partial_cmp(&other.patch) {
                    Some(Ordering::Greater) => Some(Ordering::Greater),
                    Some(Ordering::Less) => Some(Ordering::Less),
                    None => None,
                    Some(Ordering::Equal) => self.ext.partial_cmp(&other.ext),
                },
            },
        }
    }
}

impl Ord for Extended {
    fn cmp(&self, other: &Self) -> Ordering {
        match self.major.cmp(&other.major) {
            Ordering::Greater => Ordering::Greater,
            Ordering::Less => Ordering::Less,
            Ordering::Equal => match self.minor.cmp(&other.minor) {
                Ordering::Greater => Ordering::Greater,
                Ordering::Less => Ordering::Less,
                Ordering::Equal => match self.patch.cmp(&other.patch) {
                    Ordering::Greater => Ordering::Greater,
                    Ordering::Less => Ordering::Less,
                    Ordering::Equal => self.ext.cmp(&other.ext),
                },
            },
        }
    }
}

impl PartialEq<SemVer> for Extended {
    fn eq(&self, other: &SemVer) -> bool {
        self.major == other.major
            && self.minor == other.minor
            && self.patch == other.patch
            && self.ext == 0
    }
}

impl PartialOrd<SemVer> for Extended {
    fn partial_cmp(&self, other: &SemVer) -> Option<Ordering> {
        match self.major.partial_cmp(&other.major) {
            Some(Ordering::Greater) => Some(Ordering::Greater),
            Some(Ordering::Less) => Some(Ordering::Less),
            None => None,
            Some(Ordering::Equal) => match self.minor.partial_cmp(&other.minor) {
                Some(Ordering::Greater) => Some(Ordering::Greater),
                Some(Ordering::Less) => Some(Ordering::Less),
                None => None,
                Some(Ordering::Equal) => match self.patch.partial_cmp(&other.patch) {
                    Some(Ordering::Greater) => Some(Ordering::Greater),
                    Some(Ordering::Less) => Some(Ordering::Less),
                    None => None,
                    Some(Ordering::Equal) => match self.ext {
                        0 => Some(Ordering::Equal),
                        _ => Some(Ordering::Greater),
                    },
                },
            },
        }
    }
}

impl PartialEq<Rapid> for Extended {
    fn eq(&self, other: &Rapid) -> bool {
        self.major == other.major && self.minor == other.minor && self.patch == 0 && self.ext == 0
    }
}

impl PartialOrd<Rapid> for Extended {
    fn partial_cmp(&self, other: &Rapid) -> Option<Ordering> {
        match self.major.partial_cmp(&other.major) {
            Some(Ordering::Greater) => Some(Ordering::Greater),
            Some(Ordering::Less) => Some(Ordering::Less),
            None => None,
            Some(Ordering::Equal) => match self.minor.partial_cmp(&other.minor) {
                Some(Ordering::Greater) => Some(Ordering::Greater),
                Some(Ordering::Less) => Some(Ordering::Less),
                None => None,
                Some(Ordering::Equal) => match (self.patch, self.ext) {
                    (0, 0) => Some(Ordering::Equal),
                    _ => Some(Ordering::Greater),
                },
            },
        }
    }
}

impl PartialEq<u32> for Extended {
    fn eq(&self, other: &u32) -> bool {
        self.major == *other && self.minor == 0 && self.patch == 0 && self.ext == 0
    }
}

impl PartialOrd<u32> for Extended {
    fn partial_cmp(&self, other: &u32) -> Option<Ordering> {
        match self.major.partial_cmp(other) {
            Some(Ordering::Greater) => Some(Ordering::Greater),
            Some(Ordering::Less) => Some(Ordering::Less),
            None => None,
            _ => match (self.minor, self.patch, self.ext) {
                (0, 0, 0) => Some(Ordering::Equal),
                _ => Some(Ordering::Greater),
            },
        }
    }
}

impl PartialEq<Extended> for u32 {
    fn eq(&self, other: &Extended) -> bool {
        *self == other.major && other.minor == 0 && other.patch == 0 && other.ext == 0
    }
}

impl PartialOrd<Extended> for u32 {
    fn partial_cmp(&self, other: &Extended) -> Option<Ordering> {
        match other.partial_cmp(self) {
            Some(Ordering::Less) => Some(Ordering::Greater),
            Some(Ordering::Greater) => Some(Ordering::Less),
            Some(Ordering::Equal) => Some(Ordering::Equal),
            None => None,
        }
    }
}
